"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const childProcess = require("child_process");
const crypto = require("crypto");
const fs = require("fs");
const https = require("https");
const path = require("path");
const zlib = require("zlib");
const mixed_1 = require("../helpers/mixed");
const parse_1 = require("../helpers/parse");
const EMPTY_STDERR = mixed_1.dim("(empty stderr)");
async function download(cwd, env, logger) {
    var _a;
    const parseResult = parse_1.findReadAndParseElmToolingJson(cwd, env);
    switch (parseResult.tag) {
        case "ElmToolingJsonNotFound":
            logger.error(parseResult.message);
            return { tag: "Exit", statusCode: 1 };
        case "ReadAsJsonObjectError":
            logger.error(mixed_1.bold(parseResult.elmToolingJsonPath));
            logger.error(parseResult.message);
            return { tag: "Exit", statusCode: 1 };
        case "Parsed": {
            switch ((_a = parseResult.tools) === null || _a === void 0 ? void 0 : _a.tag) {
                case undefined:
                    logger.log(mixed_1.bold(parseResult.elmToolingJsonPath));
                    logger.log(`The "tools" field is missing. Nothing to download.`);
                    return { tag: "Exit", statusCode: 0 };
                case "Error":
                    logger.error(mixed_1.bold(parseResult.elmToolingJsonPath));
                    logger.error("");
                    logger.error(parse_1.printFieldErrors(parseResult.tools.errors));
                    logger.error("");
                    logger.error(mixed_1.elmToolingJsonDocumentationLink);
                    return { tag: "Exit", statusCode: 1 };
                case "Parsed":
                    logger.log(mixed_1.bold(parseResult.elmToolingJsonPath));
                    return await downloadTools(logger, parseResult.tools.parsed);
            }
        }
    }
}
exports.default = download;
async function downloadTools(logger, tools) {
    if (tools.existing.length === 0 && tools.missing.length === 0) {
        logger.log(`The "tools" field is empty. Nothing to download.`);
        return { tag: "Exit", statusCode: 0 };
    }
    for (const tool of tools.existing) {
        logger.log(`${mixed_1.bold(`${tool.name} ${tool.version}`)} already exists: ${mixed_1.dim(tool.absolutePath)}`);
    }
    if (tools.missing.length === 0) {
        return { tag: "Success", tools: tools.existing };
    }
    for (const tool of tools.missing) {
        fs.mkdirSync(path.dirname(tool.absolutePath), { recursive: true });
    }
    const toolsProgress = tools.missing.map(() => 0);
    const firstDrawTime = Date.now();
    const redraw = ({ force = false } = {}) => {
        // Without this time thing, you’ll see the progress go up to 100% and then
        // back to 0% again when using curl. It reports the progress per request –
        // even redirects. Hopefully a redirect response usually finishes in 1 second.
        if (Date.now() - firstDrawTime > 1000 || force) {
            logger.progress(tools.missing
                .map((tool, index) => {
                const progress = toolsProgress[index];
                const progressString = typeof progress === "string"
                    ? progress.padEnd(4)
                    : Math.round(progress * 100)
                        .toString()
                        .padStart(3) + "%";
                return `${mixed_1.bold(progressString)} ${tool.name} ${tool.version}`;
            })
                .join("\n"));
        }
    };
    redraw({ force: true });
    const results = await Promise.all(tools.missing.map((tool, index) => downloadAndExtract(tool, (percentage) => {
        toolsProgress[index] = percentage;
        redraw();
    }).catch((error) => {
        toolsProgress[index] = "ERR!";
        redraw();
        return new Error(downloadAndExtractError(tool, error));
    })));
    const downloadErrors = results.flatMap((result) => result instanceof Error ? result : []);
    redraw({ force: true });
    if (downloadErrors.length > 0) {
        logger.error("");
        logger.error([
            mixed_1.printNumErrors(downloadErrors.length),
            ...downloadErrors.map((error) => error.message),
        ].join("\n\n"));
        return { tag: "Exit", statusCode: 1 };
    }
    return {
        tag: "Success",
        tools: [...tools.existing, ...tools.missing],
    };
}
function downloadAndExtract(tool, onProgress) {
    return new Promise((resolve, reject) => {
        const removeExtractedAndReject = (error) => {
            try {
                hash.destroy();
                extract.destroy();
                download.kill();
                fs.unlinkSync(tool.absolutePath);
                reject(error);
            }
            catch (removeErrorAny) {
                const removeError = removeErrorAny;
                if (removeError.code === "ENOENT") {
                    reject(error);
                }
                else {
                    reject(new Error(`${error.message}\n\n${removeError.message}`));
                }
            }
        };
        const hash = crypto.createHash("sha256");
        const extract = extractFile({
            assetType: tool.asset.type,
            file: tool.absolutePath,
            onError: removeExtractedAndReject,
            onSuccess: resolve,
        });
        const download = downloadFile(tool.asset.url, {
            onData: (chunk) => {
                hash.update(chunk);
                extract.write(chunk);
            },
            onProgress,
            onError: removeExtractedAndReject,
            onSuccess: () => {
                const digest = hash.digest("hex");
                if (digest === tool.asset.hash) {
                    extract.end();
                }
                else {
                    removeExtractedAndReject(new Error(hashMismatch(digest, tool.asset.hash)));
                }
            },
        });
    });
}
function downloadAndExtractError(tool, error) {
    return `
${mixed_1.bold(`${tool.name} ${tool.version}`)}
${mixed_1.indent(`
${mixed_1.dim(`< ${tool.asset.url}`)}
${mixed_1.dim(`> ${tool.absolutePath}`)}
${error.message}
  `.trim())}
  `.trim();
}
function hashMismatch(actual, expected) {
    return `
The downloaded file does not have the expected hash!
Expected: ${expected}
Actual:   ${actual}
  `.trim();
}
function downloadFile(url, { onData, onProgress, onError, onSuccess, }) {
    let stderr = "";
    const errored = [];
    const onStderr = (chunk) => {
        stderr += chunk.toString();
        // Extract progress percentage from curl/wget.
        const matches = stderr.match(/\d+(?:[.,]\d+)?%/g) || [];
        if (matches.length > 0) {
            onProgress(Math.min(1, parseFloat(matches[matches.length - 1]) / 100));
        }
    };
    const onClose = (commandName) => (code) => {
        if (errored.includes(commandName)) {
            return;
        }
        else if (code === 0) {
            onSuccess();
        }
        else {
            onError(new Error(`${commandName} exited with non-zero exit code ${code}:\n${stderr
                .trim()
                // Remove curl’s progress bar remnants.
                .replace(/^[\s#O=-]+/g, "") || EMPTY_STDERR}`));
        }
    };
    const curl = childProcess.spawn("curl", ["-#fL", url]);
    let toKill = curl;
    curl.stdout.on("data", onData);
    curl.stderr.on("data", onStderr);
    curl.on("close", onClose("curl"));
    curl.on("error", (error) => {
        errored.push("curl");
        if (error.code === "ENOENT") {
            const wget = childProcess.spawn("wget", ["-O", "-", url]);
            toKill = wget;
            wget.stdout.on("data", onData);
            wget.stderr.on("data", onStderr);
            wget.on("close", onClose("wget"));
            wget.on("error", (error) => {
                errored.push("wget");
                if (error.code === "ENOENT") {
                    toKill = downloadFileNative(url, {
                        onData,
                        onProgress,
                        onError,
                        onSuccess,
                    });
                }
                else {
                    onError(error);
                }
            });
        }
        else {
            onError(error);
        }
    });
    return { kill: () => toKill.kill() };
}
function downloadFileNative(url, { onData, onProgress, onError, onSuccess, }, maxRedirects = 50 // This is curl’s default.
) {
    let toKill = { kill: () => request.destroy() };
    const request = https.get(url, (response) => {
        var _a;
        switch (response.statusCode) {
            case 302: {
                const redirect = response.headers.location;
                if (redirect === undefined) {
                    onError(new Error(`Got 302 without location header.`));
                }
                else if (maxRedirects <= 0) {
                    onError(new Error(`Too many redirects.`));
                }
                else {
                    toKill = downloadFileNative(redirect, {
                        onData,
                        onProgress,
                        onError,
                        onSuccess,
                    }, maxRedirects - 1);
                }
                break;
            }
            case 200: {
                const contentLength = parseInt(response.headers["content-length"] || "", 10);
                let length = 0;
                response.on("data", (chunk) => {
                    length += chunk.length;
                    onData(chunk);
                    if (Number.isFinite(contentLength) && contentLength > 0) {
                        onProgress(Math.min(1, length / contentLength));
                    }
                });
                response.on("end", onSuccess);
                break;
            }
            default:
                onError(new Error(`Unexpected status code: ${(_a = response.statusCode) !== null && _a !== void 0 ? _a : "unknown"}`));
                break;
        }
    });
    request.on("error", onError);
    return { kill: () => toKill.kill() };
}
function extractFile({ assetType, file, onError, onSuccess, }) {
    switch (assetType) {
        case "gz": {
            const gunzip = zlib.createGunzip();
            const write = fs.createWriteStream(file);
            gunzip.on("error", onError);
            write.on("error", onError);
            write.on("close", () => {
                fs.chmod(file, mixed_1.EXECUTABLE, (error) => {
                    if (error === null) {
                        onSuccess();
                    }
                    else {
                        onError(error);
                    }
                });
            });
            gunzip.pipe(write);
            return gunzip;
        }
        case "tgz":
            return extractTar({ input: "-", file, onError, onSuccess });
        // GNU tar does not support zip files, but only Windows uses zip files and
        // Windows comes with BSD tar which does support them. This could have used
        // the exact same code as for `tgz`, but it somehow fails on Windows:
        // https://stackoverflow.com/questions/63783342/windows-using-tar-to-unzip-zip-from-stdin-works-in-terminal-but-not-in-node-js
        // Workaround: Save the zip to disk, extract it and remove the zip again.
        case "zip": {
            const temp = `${file}.zip`;
            const write = fs.createWriteStream(temp);
            let toDestroy = write;
            let cleanup = () => {
                // If the caller runs `.destroy()` after we’ve already run `cleanup`,
                // don’t run the cleanup procedure again: If the cleanup succeeded
                // there’s nothing to clean; if it failed, running it again will just
                // fail again.
                cleanup = () => undefined;
                try {
                    fs.unlinkSync(temp);
                    return undefined;
                }
                catch (errorAny) {
                    const error = errorAny;
                    return error.code === "ENOENT" ? undefined : error;
                }
            };
            write.on("error", onError);
            write.on("close", () => {
                toDestroy = extractTar({
                    input: temp,
                    file,
                    onError: (error) => {
                        const cleanupError = cleanup();
                        onError(cleanupError === undefined
                            ? error
                            : new Error(`${error.message}\n\n${cleanupError.message}`));
                    },
                    onSuccess: () => {
                        const cleanupError = cleanup();
                        if (cleanupError === undefined) {
                            onSuccess();
                        }
                        else {
                            onError(cleanupError);
                        }
                    },
                });
            });
            return {
                destroy: () => {
                    toDestroy.destroy();
                    const cleanupError = cleanup();
                    if (cleanupError !== undefined) {
                        // If cleanup fails, throw the error just like `.destroy()` can.
                        throw cleanupError;
                    }
                },
                write: (chunk) => write.write(chunk),
                end: () => write.end(),
            };
        }
    }
}
function extractTar({ input, file, onError, onSuccess, }) {
    const tar = childProcess.spawn("tar", [
        "zxf",
        input,
        "-C",
        path.dirname(file),
        path.basename(file),
    ]);
    let stderr = "";
    tar.on("error", (error) => {
        if (error.code === "ENOENT") {
            onError(new Error(`tar must be installed on your system and be in ${parse_1.isWindows ? "%PATH%" : "$PATH"}:\n${error.message}`));
        }
        else {
            onError(error);
        }
    });
    tar.stderr.on("data", (chunk) => {
        stderr += chunk.toString();
    });
    tar.on("close", (code) => {
        if (code === 0) {
            onSuccess();
        }
        else {
            onError(new Error(`tar exited with non-zero exit code ${code}:\n${stderr.trim() || EMPTY_STDERR}`));
        }
    });
    return {
        destroy: () => tar.kill(),
        write: (chunk) => tar.stdin.write(chunk),
        end: () => tar.stdin.end(),
    };
}
