"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.makePs1Script = exports.makeCmdScript = exports.makeShScript = void 0;
const fs = require("fs");
const path = require("path");
const mixed_1 = require("../helpers/mixed");
const parse_1 = require("../helpers/parse");
const download_1 = require("./download");
async function postinstall(cwd, env, logger) {
    if ("NO_ELM_TOOLING_POSTINSTALL" in env) {
        return 0;
    }
    const result = await download_1.default(cwd, env, logger);
    switch (result.tag) {
        case "Exit":
            return result.statusCode;
        case "Success":
            return linkTools(cwd, logger, result.tools);
    }
}
exports.default = postinstall;
function linkTools(cwd, logger, tools) {
    const nodeModulesPath = mixed_1.findClosest("node_modules", cwd);
    /* istanbul ignore if */
    if (nodeModulesPath === undefined) {
        logger.error("No node_modules/ folder found. Install your npm dependencies before running this script.");
        return 1;
    }
    const nodeModulesBinPath = path.join(nodeModulesPath, ".bin");
    try {
        fs.mkdirSync(nodeModulesBinPath, { recursive: true });
    }
    catch (errorAny) {
        const error = errorAny;
        logger.error(`Failed to create ${nodeModulesBinPath}:\n${error.message}`);
        return 1;
    }
    for (const tool of tools) {
        const linkPath = path.join(nodeModulesBinPath, tool.name);
        // Just like npm, these overwrite whatever links are already in
        // `node_modules/.bin/`. Most likely it’s either old links from for example
        // the `elm` npm package, or links from previous runs of this script.
        const [linkPathPresentationString, what] = parse_1.isWindows
            ? // istanbul ignore next
                [symlinkShimWindows(tool, linkPath), "shims"]
            : [symlink(tool, linkPath), "link"];
        if (linkPathPresentationString instanceof Error) {
            logger.error(linkPathPresentationString.message);
            return 1;
        }
        logger.log(`${mixed_1.bold(`${tool.name} ${tool.version}`)} ${what} created: ${mixed_1.dim(`${linkPathPresentationString} -> ${tool.absolutePath}`)}`);
    }
    return 0;
}
function symlink(tool, linkPath) {
    try {
        fs.unlinkSync(linkPath);
    }
    catch (errorAny) {
        const error = errorAny;
        if (error.code !== "ENOENT") {
            return new Error(`Failed to remove old link for ${tool.name} at ${linkPath}:\n${error.message}`);
        }
    }
    try {
        fs.symlinkSync(tool.absolutePath, linkPath);
    }
    catch (errorAny) /* istanbul ignore next */ {
        const error = errorAny;
        return new Error(`Failed to create link for ${tool.name} at ${linkPath}:\n${error.message}`);
    }
    return linkPath;
}
// istanbul ignore next
function symlinkShimWindows(tool, linkPath) {
    const shLinkPath = linkPath;
    const cmdLinkPath = `${linkPath}.cmd`;
    const ps1LinkPath = `${linkPath}.ps1`;
    const linkPathPresentationString = `${linkPath}{,.cmd,.ps1}`;
    const shScript = makeShScript(tool.absolutePath);
    const cmdScript = makeCmdScript(tool.absolutePath);
    const ps1Script = makePs1Script(tool.absolutePath);
    try {
        fs.unlinkSync(shLinkPath);
        fs.unlinkSync(cmdLinkPath);
        fs.unlinkSync(ps1LinkPath);
    }
    catch (errorAny) {
        const error = errorAny;
        if (error.code !== "ENOENT") {
            return new Error(`Failed to remove old shims for ${tool.name} at ${linkPathPresentationString}:\n${error.message}`);
        }
    }
    try {
        fs.writeFileSync(shLinkPath, shScript);
        fs.chmodSync(shLinkPath, mixed_1.EXECUTABLE);
        fs.writeFileSync(cmdLinkPath, cmdScript);
        fs.chmodSync(cmdLinkPath, mixed_1.EXECUTABLE);
        fs.writeFileSync(ps1LinkPath, ps1Script);
        fs.chmodSync(ps1LinkPath, mixed_1.EXECUTABLE);
    }
    catch (errorAny) {
        const error = errorAny;
        return new Error(`Failed to create shims for ${tool.name} at ${linkPathPresentationString}:\n${error.message}`);
    }
    return linkPathPresentationString;
}
// Windows-style paths works fine, at least in Git bash.
function makeShScript(toolAbsolutePath) {
    return lf(`
#!/bin/sh
${toolAbsolutePath
        .split(/(')/)
        .map((segment) => segment === "" ? "" : segment === "'" ? "\\'" : `'${segment}'`)
        .join("")} "$@"
`);
}
exports.makeShScript = makeShScript;
// Note: Paths on Windows cannot contain `"`.
function makeCmdScript(toolAbsolutePath) {
    return crlf(`
@ECHO off
"${toolAbsolutePath}" %*
`);
}
exports.makeCmdScript = makeCmdScript;
// The shebang is for PowerShell on unix: https://github.com/npm/cmd-shim/pull/34
function makePs1Script(toolAbsolutePath) {
    return lf(`
#!/usr/bin/env pwsh
& '${toolAbsolutePath.replace(/'/g, "''")}' $args
`);
}
exports.makePs1Script = makePs1Script;
function lf(string) {
    return `${string.trim()}\n`;
}
function crlf(string) {
    return lf(string).replace(/\n/g, "\r\n");
}
